/*
    Fichier chargemail.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-de.txt , licence-en.txt , licence-es.txt ,
    licence-it.txt , licence-nl.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://libremail.tuxfamily.org/voir.php?page=droits


    Ce programme tlcharge les mail en attente sur un serveur.
    Un fichier de configuration est utilis pour se connecter  la
    boite aux lettres et pour fixer le rpertoire racine du systme
    de messagerie.

    Les mail entrants sont stocks dans le sous rpertoire "entree"
    de ce rpertoire racine et se distinguent par leur numro.

    On peut choisir de tlcharger :
    - tous les mails (option par dfaut)
    - un mail particulier identifi par son numro
*/

#define appli   // pour la dclaration de variables globales  l'application

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include "messages.h"
#include "buflect.h"
#include "ficonf.h"
#include "pop.h"
#include "numail.h"
#include "szchemin.h"


/* constantes pour indiquer si l'on efface le(s) mail(s) tlcharg(s) */
#define OPTDEFAUT  0
#define EFFACE     1
#define CONSERVE   2


/* prototype */
void chargemail (int numes, long numail, int effacement);


/* variable globale au source
   (pour viter des tonnes de passages de paramtres) */

char dirmails [szchemin]; // rpertoire de destination des messages

// chaine de caractres mmorise pour viter un appel rptitif  message ()
char mess_telech [50]; // message signalant le tlchargement d'un mail


/* programme principal */

int main (int nbarg, char *varg[])
{
    FILE *fconf;       // descripteur du fichier de configuration
    int  numes, nbmes; // numro du mail courant et nombre de mails
    long numail;       // numro du mail dans le fichier tlcharg
    int  option;       // option : dtermine si on conserve les mails rcuprs


    // rcupration du nom de l'excutable
    memcom (*varg);

    // se positionner sur le premier argument de chargemail
    varg ++;

    // rcupration de l'option de fonctionnement ventuelle
    if (--nbarg > 1 && **varg == '-')
    {
        if (varg [0][1] == 'd')
            option = EFFACE;

        else if (varg [0][1] == 'k')
            option = CONSERVE;
        else
        {
            // "Option %s incorrecte"
            aff_err_arg ("ERR_OPTION", *varg);
            nbarg = 0;   // pour afficher la syntaxe de la commande
        }

        // dcompter l'argument trait
        varg ++;
        nbarg --;
    }
    else
        // option de fonctionnement par dfaut
        option = OPTDEFAUT;

    // rcuprer le numro du mail  tlcharger s'il a t spcifi
    if (nbarg && atoi (*varg) > 0)
    {
        numes = atoi (*varg);

        // dcompter l'argument trait
        nbarg --;
        varg ++;
    }
    else
        numes = 0;  // on tlchargera tous les mails

    // controle du nombre d'arguments restant
    if (nbarg == 1)
    {
        // ouvrir le fichier de configuration
        fconf = ouvre_ficonf (*varg);

        if (fconf)
        {
            // connexion sur le compte mail du serveur pop
            if (connect_pop (fconf))
            {
                // rcuprer le rpertoire racine de la messagerie
                fgets (dirmails, sizeof (dirmails) - 7, fconf);
                dirmails [strlen (dirmails) - 1] = '\0';

                // si ce rpertoire existe
                if (access (dirmails, 0) == 0)
                {
                    // crer si ncessaire le sous repertoire d'entre
                    strcat (dirmails, "/");
                    strcat (dirmails, ficdir ("DIR_ENTREE"));
                    mkdir (dirmails, 0755);

                    // rcupration du nombre de mails
                    nbmes = nbmails ();

                    // si des mails peuvent tre rcuprs
                    if (nbmes)
                    {
                        // rcupration du numro du dernier fichier mail
                        numail = num_dermail (dirmails);

                        // Initialisation du message 
                        // afficher  chaque tlchargement
                        // "\rTlchargement du mail n %d"
                        strcpy (mess_telech, message ("TELECH_MAIL"));

                        // si demande de tlchargement d'un mail particulier
                        if (numes)
                        {
                            // si un mail existe avec ce numro
                            if (numes <= nbmes)
                            {
                                // par dfaut, lorsqu'on rcupre un mail
                                // unique, on le laisse sur le serveur
                                if (option == OPTDEFAUT)
                                    option = CONSERVE;

                                // tlchargement du mail demand
                                chargemail (numes, ++numail, option);
                            }
                            // sinon, afficher un message d'erreur
                            else
                                // "Numro de mail trop grand"
                                affiche_err ("NUMAIL_TROP_GRAND");
                        }
                        // sinon, on tlchargera tous les mails
                        else
                        {
                            // par dfaut, lorsqu'on rcupre tous les mails,
                            // on les supprime du serveur
                            if (option == OPTDEFAUT)
                                option = EFFACE;

                            // tlchargement de tous les mails
                            for (numes = 1; numes <= nbmes; numes++)
                                chargemail (numes, ++numail, option);
                        }

                        // pour l'affichage
                        putchar ('\n');

                        // sauvegarde du numro du dernier fichier mail
                        sauv_num_dermail (dirmails, numail);
                    }
                }
                else
                    // "rpertoire de messagerie %s inexistant"
                    aff_err_arg ("REP_RACINE_ABSENT", dirmails);

                // se dconnecter proprement du serveur pop
                deconnect_pop ();
            }

            // on n'a plus besoin du fichier de configuration
            fclose (fconf);
        }
    }
    else
        // "Syntaxe : %s [-(k|d)] [numro_mail] fichier_configuration"
        psyntaxe ("SYNT_CHARGEMAIL");

    // pour faire plaisir  gcc qui veut une fonction main de type int
    return (0);
}


/* tlcharge le message choisi et l'enlve du serveur si c'est demand */

void chargemail (int numes, long numail, int effacement)
{
    char bufw [120];    // buffer d'envoi d'une requte
    char ficmail [sizeof (dirmails) + 9];
    FILE *fmail;

    // cration du fichier mail
    sprintf (ficmail, "%s/r%07ld.n", dirmails, numail);

    fmail = fopen (ficmail, "w");

    if (fmail)
    {
        // message de suivi du droulement
        // "\rTlchargement du mail n %d"
        printf (mess_telech, numes);
        fflush (stdout);

        // demande de lecture du message et rception de l'acquittement
        sprintf (bufw, "RETR %d", numes);
        env_pop (bufw);
        lire_pop ();

        // termin pour ce mail si erreur d'envoi cot serveur
        if (memcmp (buf_lect, "-ERR ", 5) == 0)
        {
            // Erreur serveur pour l'accs au mail
            aff_err_argnum ("ERREUR_SERVEUR", numes);
            fclose (fmail);
            unlink (ficmail);
            return;
        }

        // lire la premire ligne de l'entte du mail
        lire_pop ();

        // lecture et copie du message jusqu' une ligne limite a un .
        while (buf_lect [0] != '.' || buf_lect [1] != '\0')
        {
            // copie de la ligne lue dans le fichier mail
            fputs (buf_lect, fmail);
            fputc ('\n', fmail);     // ncessaire car supprim de buf_lect

            // et lecture de la suivante
            lire_pop ();
        }

        // fermer le fichier contenant le mail
        fclose (fmail);

        // et supprimer si ncessaire le mail du serveur
        if (effacement != CONSERVE)
        {
            sprintf (bufw, "DELE %d", numes);
            env_pop (bufw);
            lire_pop ();
        }
    }
    else
        // "Impossible d'crire dans le fichier %s"
        aff_err_arg ("IMPOS_ECR_FICH", ficmail);
}
