/*
    Fichier chargepartaille.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-de.txt , licence-en.txt , licence-es.txt ,
    licence-it.txt , licence-nl.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://libremail.tuxfamily.org/voir.php?page=droits


    Ce programme tlcharge les mail en attente sur un serveur.
    Un fichier de configuration est utilis pour se connecter  la
    boite aux lettres et pour fixer le rpertoire racine du systme
    de messagerie.

    Les mail entrants sont stocks dans le sous rpertoire "entree"
    de ce rpertoire racine et se distinguent par leur numro.

    Cette version de chargemail charge en premier les mails les plus
    petits, puis les mails de plus en plus gros.

    Contrairement  chargemail, les mails tlchargs sont dtruits du
    serveur sans attendre la fin du tlchargement. Ainsi, s'il y a
    une rupture de connexion pendant l'utilisation de chargepartaille,
    il ne sera pas ncessaire de tlcharger une autre fois la plupart
    des mails dj rcuprs.
*/

#define appli   // pour la dclaration de variables globales  l'application

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include "messages.h"
#include "buflect.h"
#include "ficonf.h"
#include "pop.h"
#include "numail.h"
#include "szchemin.h"


/* constantes */
#define nb_sec        20  // dure entre 2 validations d'effacement des mails
#define taille_dep  5000  // taille maximale des mails chargs en premier
#define szmaxliste   100  // longueur max de la liste des tailles de mail


/* prototypes */
void chargepartaille ();
void chargemail (int numes, long numail);


/* variables globales au source
   (pour viter des tonnes de passages de paramtres) */

char dirmails [szchemin]; // rpertoire de destination des messages
FILE *fconf;              // descripteur du fichier de configuration

// chaine de caractres mmorise pour viter un appel rptitif  message ()
char mess_telech [50]; // message signalant le tlchargement d'un mail


/* programme principal */

int main (int nbarg, char *varg[])
{
    // rcupration du nom de l'excutable
    memcom (*varg);

    // controle du nombre d'arguments
    if (--nbarg == 1)
    {
        // ouvrir le fichier de configuration
        fconf = ouvre_ficonf (varg [1]);

        if (fconf)
        {
            // connexion sur le compte mail du serveur pop
            if (connect_pop (fconf))
            {
                // rcuprer le rpertoire racine de la messagerie
                fgets (dirmails, sizeof (dirmails) - 7, fconf);
                dirmails [strlen (dirmails) - 1] = '\0';

                // si ce rpertoire existe
                if (access (dirmails, 0) == 0)
                {
                    // crer si ncessaire le sous repertoire d'entre
                    strcat (dirmails, "/");
                    strcat (dirmails, ficdir ("DIR_ENTREE"));
                    mkdir (dirmails, 0755);

                    // Initialisation du message 
                    // afficher  chaque tlchargement
                    // "\rTlchargement du mail n %d"
                    strcpy (mess_telech, message ("TELECH_MAIL"));

                    // tlcharger les mails par taille croissante
                    chargepartaille ();
                }
                else
                    // "rpertoire de messagerie %s inexistant"
                    aff_err_arg ("REP_RACINE_ABSENT", dirmails);

                // se dconnecter proprement du serveur pop
                deconnect_pop ();
            }

            // on n'a plus besoin du fichier de configuration
            fclose (fconf);
        }
    }
    else
        // "Syntaxe : %s fichier_configuration"
        psyntaxe ("SYNT_CHARGEPARTAIL");

    // pour faire plaisir  gcc qui veut une fonction main de type int
    return (0);
}


/* tlcharge les mails par taille croissante en validant
   priodiquement leur destruction sur le serveur */

void chargepartaille ()
{
    long listetaille [szmaxliste]; // nombre d'octets des diffrents mails
    int  posliste, szliste; // position et nombre d'lments dans listetaille
    long numail;            // numro du mail dans le fichier tlcharg
    long taillemax;         // taille maximale mails tlchargs  cette tape
    int  etape;             // numro d'tape
    unsigned long heurefin; // pour dcider de changer d'tape
    long tampon;            // variable de travail


    // initialisation
    taillemax = taille_dep;
    etape = 1;

    // rcupration du numro du dernier fichier mail dj tlcharg
    numail = num_dermail (dirmails);

    // rcupration du nombre de mails
    szliste = nbmails ();

    // tant qu'il y a des mails  tlcharger
    while (szliste > 0)
    {
        // "%d mails en attente de tlchargement\n"
        printf (message ("A_TELECHARGER"), szliste);

        // limiter si ncessaire le nombre de mails traits  cette tape
        if (szliste > szmaxliste)
            szliste = szmaxliste;

        // calcul de l'heure du changement d'tape
        heurefin = time (0) + nb_sec;

        // mmoriser les tailles des mails
        env_pop ("LIST");
        posliste = 0;

        // sauter le message d'entte de la liste
        lire_pop ();

        do
        {
            // lire la taille d'un mail
            lire_pop ();

            // s'ils sont trop nombreux, on ne mmorise que les premiers
            if (posliste < szliste)
                sscanf (buf_lect, "%ld%ld", &tampon, &listetaille [posliste++]);
        }
        while (buf_lect [0] != '.');

        // se positionner en dbut de liste
        posliste = 0;

        // rechercher le premier mail dont la taille convient
        while (listetaille [posliste] > taillemax)
        {
            // tant qu'on n'en trouve pas, on avance dans la liste
            if (++posliste == szliste)
            {
                // si pas trouv en fin de liste, essayer avec taille suprieure
                taillemax = taillemax * 2;
                posliste  = 0;
            }
        }

        // "Etape %d, mails de %ld octets maximum\n"
        printf (message ("ETAPE_TELECH"), etape++, taillemax);

        // tlchargement des mails dont la taille convient
        do
        {
            if (listetaille [posliste] <= taillemax)
                // attention  la numrotation des mails
                chargemail (posliste + 1, ++numail);
        }
        while (++posliste < szliste && heurefin > time (0));

        // pour l'affichage
        putchar ('\n');

        // valider la destruction des mails tlchargs
        deconnect_pop ();

        // et recommencer une tape similaire
        rewind (fconf);
        connect_pop (fconf);
        szliste = nbmails ();
    }

    // sauvegarde du numro du dernier fichier mail
    sauv_num_dermail (dirmails, numail);
}


/* tlcharge le mail choisi et l'enlve du serveur */

void chargemail (int numes, long numail)
{
    char bufw [120];    // buffer d'envoi d'une requte
    char ficmail [sizeof (dirmails) + 9];
    FILE *fmail;

    // cration du fichier mail
    sprintf (ficmail, "%s/r%07ld.n", dirmails, numail);

    fmail = fopen (ficmail, "w");

    if (fmail)
    {
        // message de suivi du droulement
        // "\rTlchargement du mail n %d"
        printf (mess_telech, numes);
        fflush (stdout);

        // demande de lecture du message et rception de l'acquittement
        sprintf (bufw, "RETR %d", numes);
        env_pop (bufw);
        lire_pop ();

        // termin pour ce mail si erreur d'envoi cot serveur
        if (memcmp (buf_lect, "-ERR ", 5) == 0)
        {
            // Erreur serveur pour l'accs au mail
            aff_err_argnum ("ERREUR_SERVEUR", numes);
            fclose (fmail);
            unlink (ficmail);
            return;
        }

        // lire la premire ligne de l'entte du mail
        lire_pop ();

        // lecture et copie du message jusqu' une ligne limite a un .
        while (buf_lect [0] != '.' || buf_lect [1] != '\0')
        {
            // copie de la ligne lue dans le fichier mail
            fputs (buf_lect, fmail);
            fputc ('\n', fmail);     // ncessaire car supprim de buf_lect

            // et lecture de la suivante
            lire_pop ();
        }

        // fermer le fichier contenant le mail
        fclose (fmail);

        // et supprimer si ncessaire le mail du serveur
        sprintf (bufw, "DELE %d", numes);
        env_pop (bufw);
        lire_pop ();
    }
    else
        // "Impossible d'crire dans le fichier %s"
        aff_err_arg ("IMPOS_ECR_FICH", ficmail);
}
