/*
    Fichier detruitmail.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-de.txt , licence-en.txt , licence-es.txt ,
    licence-it.txt , licence-nl.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://libremail.tuxfamily.org/voir.php?page=droits


    Ce programme permet de dtruire des mail en attente sur le
    serveur de messagerie.
    Les mails  dtruire sont identifis par leur numro.

    Un fichier de configuration est utilis pour se connecter  la
    boite aux lettres et pour fixer le rpertoire racine du systme
    de messagerie.
*/

#define appli   // pour la dclaration de variables globales  l'application

#include <stdio.h>
#include <string.h>
#include "messages.h"
#include "buflect.h"
#include "ficonf.h"
#include "pop.h"


/* prototype */
void detruitmail (int numes);


/* variable globale au source
   (pour viter des tonnes de passages de paramtres) */

// chaine de caractres mmorise pour viter un appel rptitif  message ()
char mess_detr [50]; // message signalant la destruction d'un mail


/* programme principal */

int main (int nbarg, char *varg[])
{
    FILE *fconf;   // descripteur du fichier de configuration
    int  numes;    // numro d'un mail
    int  finint;   // dernier numro d'un intervalle
    int  nbmes;    // nombre de mails sur le serveur
    char *carg1;   // caractre de l'argument 1


    // rcupration du nom de l'excutable
    memcom (*varg);

    // se positionner sur le premier argument de detruitmail
    varg ++;

    // controle du nombre d'arguments et de l'argument 1
    if (--nbarg == 2 && ((strcmp (*varg, "-a") == 0) ||
                         ('1' <= **varg && **varg <= '9')))
    {

        // ouvrir le fichier de configuration
        fconf = ouvre_ficonf (varg [1]);

        if (fconf)
        {
            // connexion sur le compte mail du serveur pop
            if (connect_pop (fconf))
            {
                // rcupration du nombre de mails
                nbmes = nbmails ();

                // si des mails peuvent tre supprims
                if (nbmes)
                {
                    // si option -a
                    if (**varg == '-')
                    {
                        // on dtruit tous les mails du serveur
                        for (numes = 1; numes <= nbmes; numes++)
                            detruitmail (numes);
                    }
                    // sinon
                    else
                    {
                        // on va analyser l'argument 1 caractre par caractre
                        carg1 = *varg;

                        while (*carg1)
                        {
                            // rcuprer un numro de mail
                            numes = 0;

                            while ('0' <= *carg1 && *carg1 <= '9')
                            {
                                numes = (numes * 10) + (*carg1 & 0x0F);
                                carg1 ++;
                            }

                            // si le sparateur qui suit est un -
                            if (*carg1 == '-')
                            {
                                // se positionner sur le caractre qui suit
                                carg1 ++;

                                // rcuprer le numro du dernier
                                // mail de l'intervalle
                                finint = 0;

                                while ('0' <= *carg1 && *carg1 <= '9')
                                {
                                    finint = (finint * 10) + (*carg1 & 0x0F);
                                    carg1 ++;
                                }

                                // si numro dernier mail trop grand
                                if (finint > nbmes)
                                {
                                    // signaler l'erreur
                                    // "Numro de mail trop grand"
                                    affiche_err ("NUMAIL_TROP_GRAND");

                                    // et corriger l'intervalle
                                    finint = nbmes;
                                }

                                // dtruire les mails de l'intervalle
                                while (numes <= finint)
                                    detruitmail (numes ++);
                            }
                            // sinon
                            else
                            {
                                // s'il existe un mail avec le numro rcupr
                                if (0 < numes && numes <= nbmes)
                                    // dtruire ce mail
                                    detruitmail (numes);

                                // sinon, afficher un message d'erreur adapt
                                else if (numes > nbmes)
                                {
                                    fprintf (stderr, "%d : ", numes);

                                    // "Numro de mail trop grand"
                                    affiche_err ("NUMAIL_TROP_GRAND");
                                }
                                else
                                    // "Un numro de mail ne peut tre gal  0"
                                    affiche_err ("NUMAIL_NUL");
                            }

                            // si on trouve le sparateur ,
                            if (*carg1 == ',')
                                // se positionner sur le numro de mail suivant
                                carg1++;

                            // sinon si non fin de chaine
                            else if (*carg1)
                            {
                                // message d'erreur
                                fprintf (stderr, message ("ERR_SEP"), *carg1);

                                // on arrte l'analyse de l'argument 1
                                *carg1 = '\0';
                            }
                        }
                    }
                }

                // se dconnecter proprement du serveur pop
                deconnect_pop ();
            }

            // on n'a plus besoin du fichier de configuration
            fclose (fconf);
        }
    }
    else
        // "Syntaxe : %s [-(k|d)] [numro_mail] fichier_configuration"
        psyntaxe ("SYNT_DETRUITMAIL");

    // pour faire plaisir  gcc qui veut une fonction main de type int
    return (0);
}


/* supprime le mail choisi du serveur de messagerie */

void detruitmail (int numail)
{
    char bufw [120];    // buffer d'envoi d'une requte


    // Destruction du mail numro
    printf (message ("DESTRUCT_MAIL"), numail);

    // dtruire le mail
    sprintf (bufw, "DELE %d", numail);
    env_pop (bufw);
    lire_pop ();
}
