/*
    Fichier envmail.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-de.txt , licence-en.txt , licence-es.txt ,
    licence-it.txt , licence-nl.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://libremail.tuxfamily.org/voir.php?page=droits


    Ce programme envoie les fichiers mail en attente d'expdition

    Un fichier de configuration est utilis pour se connecter au
    serveur smtp et pour fixer le rpertoire racine du systme
    de messagerie.

    Les mail  envoyer sont stocks dans le sous rpertoire "sortie"
    de ce rpertoire racine et se distinguent par leur numro.

    Aprs envoi, ces fichiers sont dplacs dans le sous rpertoire
    "envoyes" du systme de messagerie
*/

#define appli   // pour la dclaration de variables globales  l'application

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <dirent.h>
#include "messages.h"
#include "buflect.h"
#include "ficonf.h"
#include "smtp.h"
#include "fmail.h"
#include "szchemin.h"
#include "encodage.h"
#include "genchampdate.h"


/* prototypes */
void envmail (char *fichmail);
int  start (char *motcle);
void env_adr (char *entete, char *donnees);
void env_dest (long poschamp);
void genfromlig ();
int  carspe_dans_ligne ();
void encode_ligne (int encodage);


/* variables globales au source
   (pour viter des tonnes de passages de paramtres) */

char adr_mail [120];    // adresse mail de l'expditeur
char direnv [20];       // rpertoire des fichiers mail envoys
struct dirent *entree;  // entre de rpertoire sur fichier mail courant
int  nbdest_ok;         // nombre de destinataires du mail accepts par smtp


/* programme principal */

int main (int nbarg, char *varg[])
{
    char dirmails [szchemin]; // rpertoire racine des fichiers mail
    char serv_smtp [120];  // nom du serveur smtp utilis
    char mess_envoi [120]; // mmorise le texte  afficher  chaque envoi de mail
    FILE *fconf;           // descripteur fichier de configuration du compte mail
    DIR  *repert;          // structure pour explorer un rpertoire
    int  numail;           // numro d'ordre du fichier mail qu'on envoie
    int  i ;                // compteur


    // rcupration du nom de l'excutable
    memcom (*varg);

    // controle du nombre d'arguments
    if (--nbarg == 1)
    {
        // ouvrir le fichier de configuration
        fconf = ouvre_ficonf (varg [1]);

        if (fconf)
        {
            // on saute les 3 lignes concernant le serveur pop
            for (i = 0; i < 3; i++)
                while (fgetc (fconf) != '\n')
                    ;

            // recuprer le rpertoire racine des fichiers mails
            fgets (dirmails, szchemin, fconf);
            dirmails [strlen (dirmails) - 1] = '\0';

            // vrifier son existence en s'y positionnant
            if (chdir (dirmails) < 0)
            {
                fclose (fconf);  // pour sortie proprement
                // "Rpertoire de messagerie %s inexistant"
                errfatale ("REP_RACINE_ABSENT", dirmails);
            }

            // faire de mme avec le rpertoire d'envoi des mails
            if (chdir (ficdir ("DIR_SORTIE")) < 0)
            {
                fclose (fconf);  // pour sortie proprement
                // "Pas de rpertoire d'envoi des mails"
                errfatale ("REP_ENVOI_ABSENT", "");
            }

            // crer si ncessaire le rpertoire des mails envoys
            sprintf (direnv, "../%s", ficdir ("DIR_ENVOYES"));
            mkdir (direnv, 0755);

            // rcuprer la description de l'expditeur des mails
            fgets (adr_mail, 120, fconf);
            adr_mail [strlen (adr_mail) - 1] = '\0';

            // recuprer le nom du serveur smtp
            fgets (serv_smtp, 120, fconf);
            serv_smtp [strlen (serv_smtp) - 1] = '\0';

            // on n'a plus besoin du fichier de configuration
            fclose (fconf);

            // si on peut se connecter au serveur smtp
            if (connect_smtp (serv_smtp))
            {
                // accder au rpertoire
                repert = opendir (".");

                // lire un nom de fichier
                entree = readdir (repert);

                // initialisation compteur
                numail = 1;

                // Mmorisation du teste  afficher  chaque envoi de mail
                // "\rEnvoi du mail n %d"
                // afin d'viter d'appeller la fonction message () chaque fois
                strcpy (mess_envoi, message ("ENVOI_MAIL"));

                // tant qu'il y a des noms de fichier  lire
                while (entree)
                {
                    // si le nom du fichier courant fait 8 caractres
                    if (strlen (entree->d_name) == 8)
                    {
                        // message de suivi du droulement
                        // "\rEnvoi du mail n %d"
                        printf (mess_envoi, numail++);
                        fflush (stdout);

                        // envoyer le mail qu'il contient
                        envmail (entree->d_name);
                    }

                    // et lire le nom du fichier suivant
                    entree = readdir (repert);
                }

                // pour l'affichage
                putchar ('\n');

                // fermeture de la connexion smtp
                deconnect_smtp ();
            }
        }
    }
    else
        // "Syntaxe : %s fichier_configuration"
        psyntaxe ("SYNT_GENE_FICONF");

    // pour faire plaisir  gcc qui veut une fonction main de type int
    return (0);
}



/* envoi du fichier mail pass en paramtre */

void envmail (char *fichmail)
{
    char nouvfic [18];  // chemin d'accs relatif au fichier mail aprs transfert
    char bufFrom [120]; // contenu du champ expditeur du message
    char bufw [120];    // pour gnrer les lignes User-agent et peut tre Date
    long pos_deblig;    // position dans le fichier en dbut de ligne
    long posTo, posCc, posBcc; // position champs des destinataires
    int  Useagent;      // simple indicateur
    int  Usedate;       // simple indicateur
    int  encodage;      // jeu de caractres d'une ligne d'entte


    // ouvrir le fichier mail  envoyer
    fmail = fopen (fichmail, "r");

    // si ouverture russie
    if (fmail)
    {
        // initialisation
        *bufFrom = '\0';
        posTo    =  -1;
        posCc    =  -1;
        posBcc   =  -1;
        Useagent =   0;
        Usedate  =   0;

        // lecture de l'entte et mmorisation
        // de la position des champs importants
        do
        {
            // rcuprer la position en dbut de ligne
            pos_deblig = ftell (fmail);

            // lire une ligne de l'entte du message
            lire_fmail ();

            // reprage des champs importants et mmorisation de leur position
            if (start ("From"))
            {
                buf_lect [120] = '\0';  // troncation si ncessaire
                strcpy (bufFrom, buf_lect);  // on mmorise le champ expditeur
            }
            else if (start ("To"))
                posTo = pos_deblig;
            else if (start ("Cc"))
                posCc = pos_deblig;
            else if (start ("Bcc"))
                posBcc = pos_deblig;
            else if (start ("X-Mailer") || start ("User-Agent"))
                Useagent = 1;
            else if (start ("Date"))
                Usedate = 1;
        }
        while (buf_lect [0] != '\0');  // lecture entte termine si ligne vide

        // expditeur du message
        if (*bufFrom)
            env_adr ("MAIL FROM", bufFrom);
        else
            env_adr ("MAIL FROM", adr_mail);

        // destinataires du message
        // initialisation compteur
        nbdest_ok = 0;

        // on dclare les destinataires un par un
        if (posTo >= 0)
            env_dest (posTo);

        if (posCc >= 0)
            env_dest (posCc);

        if (posBcc >= 0)
            env_dest (posBcc);

        // si au moins un destinataire accept
        if (nbdest_ok)
        {
            // prparation de l'envoi du message
            envoie_smtp ("DATA");
            lire_smtp ();

            // on revient au dbut de l'entte du mail
            rewind (fmail);

            // si expditeur manquant, gnrer la ligne From
            if (*bufFrom == 0)
                genfromlig ();
            // sinon, lire la premre ligne de l'entte
            else
                lire_fmail ();

            // rajouter la date si elle n'tait pas dans l'entte du mail
            if (! Usedate)
            {
                gen_champ_date (bufw);
                envoie_smtp (bufw);
            }

            // copie de l'entte
            while (buf_lect [0] != '\0')
            {
                // s'il y a des destinataires en copie cache
                if (start ("Bcc"))
                {
                    // on le signale par un message gnrique
                    envoie_smtp ("Bcc: (...)");

                    // et on saute ces destinataires
                    do
                        lire_fmail ();
                    while (*buf_lect == ' ' || *buf_lect == '\t');
                }
                // cas gnral
                else
                {
                    // si la ligne contient des caractres spciaux
                    encodage = carspe_dans_ligne ();

                    if (encodage)
                    {
                        // encoder cette ligne
                        encode_ligne (encodage);
                    }

                    // on envoie la ligne lue
                    envoie_smtp (buf_lect);

                    // et on lit la suivante
                    lire_fmail ();
                }
            }

            // si pas de nom de User-Agent dans l'entte envoye
            if (! Useagent)
            {
                // le rajouter, en plus, ce n'est pas Outlook :-))
                // "libremail : logiciel libre multilingue"
                sprintf (bufw, "User-Agent: libremail : %s\n",
                                message ("QUALIF_LIBREMAIL"));
                envoie_smtp (bufw);
            }

            // recopier le message
            do
            {
                // si la ligne ne contient qu'un  .
                if (buf_lect [0] == '.' && ! buf_lect [1])
                    // le doubler pour ne pas faire croire  une fin de message
                    envoie_smtp ("..");
                else
                    // sinon, envoyer la ligne telle quelle
                    envoie_smtp (buf_lect);
            }
            while (lire_fmail ());  // on s'arrtera en fin de fichier mail

            // message indiquant la fin du mail
            envoie_smtp (".");
        }
        // sinon
        else
            // annulation de l'envoi du message
            envoie_smtp ("RSET");

        // accus de rception smtp
        lire_smtp ();

        // transfert du mail termin
        fclose (fmail);

        // on peut dplacer ce fichier
        sprintf (nouvfic, "%s/%s", direnv, fichmail);
        rename (fichmail, nouvfic);
    }
    else
        // "\nImpossible de lire le fichier %s"
        aff_err_arg ("IMPOS_LECT_FICH", fichmail);
}



/* cherche si la dernire ligne lue commence par un mot cl particulier */

int start (char *motcle)
{
    int i;

    i = 0;

    // on teste caractre par caractre en ignorant la casse
    while (motcle [i] && tolower (buf_lect [i]) == tolower (motcle [i]))
        i++;

    // le mot cl doit tre suivi de  :  dans la ligne lue
    return (buf_lect [i] == ':');
}



/* gnre une ligne dclarant l'expditeur ou des destinataires du message */

void env_adr (char *entete, char *ligne)
{
    char bufw [sz_buflect];  // buffer d'envoi d'une commande smtp
    int  i, j;  // compteurs


    // initialisation : positionnement en dbut de ligne
    i = 0;

    // rechercher l'adresse dans la ligne
    while (ligne [i] && ligne [i] != '<' && ligne [i] != '@')
    {
        if (! ligne [++i])
        {
            // "Adresse %s errone"
            aff_err_arg ("ERR_ADR_DEST", ligne);
            return;
        }
    }

    // se positionner au dbut de l'adresse
    while (i >= 0 && ligne [i] != '<' && ligne [i] != ' ' && ligne [i] != '\t')
        i--;

    i++;   // on sera sur le 1er caractre de l'adresse


    // dbut de la commande smtp  envoyer
    sprintf (bufw, "%s: <", entete);
    j = strlen (bufw);

    // copie de l'adresse
    do
        bufw [j++] = ligne [i++];
    while (ligne [i] && ligne [i] != '>' && ligne [i] != ','
                     && ligne [i] != ' ' && ligne [i] != '\t');

    // terminaison de la chaine  envoyer
    bufw [j++] = '>';
    bufw [j]   = '\0';

    // envoi de la commande smtp
    envoie_smtp (bufw);

    // lire l'accus de rception
    lire_smtp ();

    // si on traite les destinataires du mail
    if (*entete == 'R')
    {
        // si destinataire accept
        if (buf_lect [0] == '2' && buf_lect [1] == '5')
            // le comptabiliser
            nbdest_ok ++;
        else
        {
            // sinon, affichage de l'adresse  problme
            // "\nFichier %s, "
            // "Destinataire %s refus\n"
            printf (message ("NOM_FICH"), entree->d_name);
            printf (message ("REFUS_DEST"), bufw + 9);
            puts (buf_lect);
        }

        // dans les 2 cas, passer  l'adresse suivante
        if (ligne [i] == '>')
            i++;

        while (ligne [i] == ' ' || ligne [i] == ',')
            i++;

        // et traiter les autres adresses de la ligne
        if (ligne [i])
            env_adr (entete, ligne + i);
    }
}



// dclare les destinataires du message contenu dans un champ
// To: , Cc ou Bcc .

void env_dest (long poschamp)
{
    char lignedest [sz_buflect];  // ligne contenant des destinataires


    // se positionner sur la premire ligne du champ destinataires  analyser
    fseek (fmail, poschamp, SEEK_SET);
    lire_fmail ();

    // tant qu'on n'a pas trait toutes les lignes du champ
    do
    {
        // copie ncessaire car la rponse  une requte smtp crasera la
        // valeur de buf_lect qui pouvait contenir plusieurs destinataires
        strcpy (lignedest, buf_lect);

        // dclarer les adresses de destinataires contenus dans la ligne lue
        env_adr ("RCPT TO", lignedest);

        // passer  la ligne suivante
        lire_fmail ();
    }
    while (*buf_lect == ' ' || *buf_lect == '\t');
}



/* gnre la ligne From de l'entte du mail  envoyer
   si elle ne figure pas dans le fichier mail */

void genfromlig ()
{
    int  i, j;  // compteurs


    // initialisation
    strcpy (buf_lect, "From: ");
    i = 0;
    j = 6;

    // recopier le dbut de la description de l'expditeur
    while (adr_mail [i] && adr_mail [i] != '<' && adr_mail [i] != '@')
    {
        buf_lect [j++] = adr_mail [i];

        // termin si pas de @ dans l'adresse de l'expditeur
        if (! adr_mail [i++])
        {
            // "Adresse Email errone dans le fichier de configuration"
            affiche_msg ("ERR_ADR_MAIL");
            return;
        }
    }

    // se positionner au dbut de l'adresse
    while (i >= 0 && adr_mail [i] != '<' && adr_mail [i] != ' '
                                        && adr_mail [i] != '\t')
    {
        i--;
        j--;
    }

    // gnrer un espace si ncessaire
    if (adr_mail [i] != '<')
        buf_lect [j++] = ' ';

    // et le < de dbut d'adresse
    buf_lect [j++] = '<';

    // aller sur le 1er caractre de l'adresse
    i++;

    // recopier l'adresse
    do
        buf_lect [j++] = adr_mail [i++];
    while (adr_mail [i] && adr_mail [i] != '>' && adr_mail [i] != ' '
                                               && adr_mail [i] != '\t');

    // terminaison de l'adresse
    buf_lect [j++] = '>';
    buf_lect [j]  = '\0';
}



/* vrifie s'il y a au moins un caractre spcial  encoder
   dans la ligne d'entte et dtermine le mode d'encodage */

int carspe_dans_ligne ()
{
    int i;  // compteur


    // initialisation
    i = 0;

    // recherche du premier caractre spcial
    while (buf_lect [i])
    {
        // si caractre spcial trouv
        if (buf_lect [i++] & 0x80)
        {
            // dtermination du jeu de caractres utilis

            // squence UTF-8 de 2 caractres
            // on refuse les squences commenant par 0xC0 ou 0xC1
            // interdites par UTF-8 et qui ont pos des problmes de scurit
            if (0xC2 <= buf_lect [i-1] && buf_lect [i-1] <= 0xDF
             && 0x80 <= buf_lect [i]   && buf_lect [i]   <= 0xBF)
                // on a trouv une squence de caractres encode UTF-8
                return (Utf8);

            // squence UTF-8 de 3 caractres
            // mme remarque pour les squences commenant par 0xE0
            else if (0xE1 <= buf_lect [i-1] // && buf_lect [i-1] <= 0xFF
                  && 0x80 <= buf_lect [i]   && buf_lect [i]   <= 0xBF
                  && 0x80 <= buf_lect [i+1] && buf_lect [i+1] <= 0xBF)
                // on a trouv une squence de caractres encode UTF-8
                return (Utf8);

            // on pourrait dtecter de mme des squences
            // UTF-8 de 4 caractres mais est-ce bien utile ?

            else
                // on considre que le jeu de caractres utilis est ISO-8859-n
                return (Iso8859);
        }
    }

    // aucun caractre spcial trouv
    return (0);
}



/* encode une ligne d'entte contenant des caractres spciaux */

void encode_ligne (int encodage)
{
    unsigned char lignecopie [sz_buflect];  // copie de buf_lect avant encodage
    int  dans_encode; // indicateur : encodage du mot courant
    int  i, j;        // compteurs


    // se positionner aprs le premier blanc ou la premire tabulation
    j = 0;

    while (buf_lect [j] != ' ' && buf_lect [j] != '\t')
        j++;

    j++;

    // recopier le reste de la ligne
    strcpy (lignecopie, buf_lect + j);

    // si ligne Subject:
    if (start ("Subject"))
    {
        // l'encodage va porter sur tout le reste de la ligne
        if (encodage == Iso8859)
            strcpy (buf_lect + j, "=?ISO-8859-15?Q?");
        else
            strcpy (buf_lect + j, "=?UTF-8?Q?");

        // j indique le premier caractre libre de buf_lect
        j = strlen (buf_lect);

        for (i = 0; lignecopie [i] != '\0'; i++)
        {
            // les blancs et caractres de controle sont remplacs par des _
            if (lignecopie [i] <= ' ')
                buf_lect [j++] = '_';

            // encodage des caractres spciaux
            else if (lignecopie [i] >= 0x7E || lignecopie [i] == '='
                  || lignecopie [i] == '?' || lignecopie [i] == '_')
            {
                sprintf (buf_lect + j, "=%02X", lignecopie [i]);

                // tenir compte du nombre de caractres rajouts
                j = j + 3;
            }
            // et simple copie des autres caractres
            else
                buf_lect [j++] = lignecopie [i];
        }

        // on termine la ligne
        strcpy (buf_lect + j, "?=");
    }
    // sinon, on encodera par mot  partir du premier caractre spcial
    else
    {
        // initialisation
        dans_encode = 0;

        for (i = 0; lignecopie [i] != '\0'; i++)
        {
            // si fin de mot ou caractre de controle (indsirable)
            if (lignecopie [i] <= ' ')
            {
                // si encodage en cours
                if (dans_encode)
                {
                    // terminer l'encodage
                    buf_lect [j++] = '?';
                    buf_lect [j++] = '=';
                    dans_encode = 0;
                }

                // recopier le ' '
                // les caractres de controle sont aussi remplacs par des ' '
                buf_lect [j++] = ' ';
            }

            // sinon, si caractre spcial  encoder
            else if (lignecopie [i] >= 0x7E || lignecopie [i] == '='
                  || lignecopie [i] == '?' || lignecopie [i] == '_')
            {
                // si pas d'encodage en cours
                if (! dans_encode)
                {
                    // le dmarrer
                    if (encodage == Iso8859)
                        strcpy (buf_lect + j, "=?ISO-8859-15?Q?");
                    else
                        strcpy (buf_lect + j, "=?UTF-8?Q?");

                    j = strlen (buf_lect);
                    dans_encode = 1;
                }

                // encoder le caractre
                sprintf (buf_lect + j, "=%02X", lignecopie [i]);

                // tenir compte du nombre de caractres rajouts
                j = j + 3;
            }

            // simple copie des autres caractres
            else
                buf_lect [j++] = lignecopie [i];
        }

        // si encodage en cours
        if (dans_encode)
        {
            // le terminer
            buf_lect [j++] = '?';
            buf_lect [j++] = '=';
        }

        // terminer la ligne
        buf_lect [j] = '\0';
    }
}
