/*
    Fichier genentete.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-de.txt , licence-en.txt , licence-es.txt ,
    licence-it.txt , licence-nl.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://libremail.tuxfamily.org/voir.php?page=droits


    Contient diffrentes fonctions pour :

    - insrer l'adresse de l'expditeur du mail qu'on cre
    - slectionner et insrer les destinataires
    - dater le mail
    - numroter les mails  envoyer
*/

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include "messages.h"
#include "buflect.h"
#include "szchemin.h"
#include "genentete.h"
#include "modepage.h"
#include "carspe.h"


#define szbufadr   120   // longueur max d'une adresse dans le carnet
#define octet      unsigned char


/* variables globales au source
   (pour viter des tonnes de passages de paramtres) */

static char  **adresse;   // les adresses de destinataires
static octet *selection;  // indique si le destinataire a t slectionn


/* recupre le nom du rpertoire des mails  envoyer */

void recupdirenv (int nbarg, char *arg2)
{
    // rcuprer le nom du rpertoire des fichiers mail  envoyer
    if (getenv ("mailenv"))
        // rcupration dans variable environnement cre par vmaildir
        strcpy (direnv, getenv ("mailenv"));
    else
    {
        if (nbarg == 2)
            // rcupration dans 2me argument pass  la commande
            strcpy (direnv, arg2);
        else
        {
            // il ne reste plus que la saisie au clavier
            // "Nom du rpertoire racine des mails : "
            affiche_msg_nocr ("REPERT_RACINE");
            fgets (direnv, szchemin, stdin);

            //  cause du \n retourn par fgets
            direnv [strlen (direnv) - 1] = '\0';
        }

        // si nom du rpertoire racine correct
        if (access (direnv, 0) == 0)
        {
            // fabriquer le nom du rpertoire des mails  envoyer
            strcat (direnv, "/");
            strcat (direnv, ficdir ("DIR_SORTIE"));
        }
        else
            // sinon, termin sur erreur
            // "Rpertoire %s inexistant"
            errfatale ("REPERT_INEXISTANT", direnv);
    }

    // crer le rpertoire d'envoi des mails si ncessaire
    mkdir (direnv, 0755);

    // s'il existe, on continue
    if (access (direnv, 0) < 0)
        // "Le rpertoire %s n'a pu tre cr"
        errfatale ("REPERT_NON_CREE", direnv);
}



/* rajoute un champ Message-Id  l'entte du mail */

void gen_mes_id (FILE *ficentete, long numail, char *adr_exped)
{
    int i;  // compteur


    // parcour de l'adresse de l'expditeur
    i = 0;

    // recherche du champ adresse
    while (adr_exped [i] && adr_exped [i] != '@')
        i++;

    // on se positionne au dbut de l'adresse
    while (i >= 0 && adr_exped [i] != '<' && adr_exped [i] != ' ')
        i--;

    // gnrer la ligne Message-Id
    fprintf (ficentete, "Message-Id: <%ld-%s>\n", numail, adr_exped + i + 1);
}



/* rajoute la date et l'heure courante  la fin de l'entte du mail

   Le champ Date: gnr tient compte du rglage le plus frquent sur
   les ordinateurs europens : l'ordinateur est rgl sur l'heure locale.
 */

void ajout_date (char *nomfic)
{
    char commande [szchemin + 42];  // la commande que l'on va excuter

    // revenir  la langue de Bill Gates utilise pour les mails
    unsetenv ("LANG");
    unsetenv ("LANGUAGE");
    unsetenv ("LC_TIME");
    unsetenv ("LC_ALL");

    // rajouter la date avec le bon format dans le fichier
    // le fuseau horaire est indiqu en chiffres, mais pas en lettres
    sprintf (commande, "date \"+%s\" >> %s", "Date: %a, %d %b %Y %T %z",nomfic);
    system (commande);
}



/* runit les 2 morceaux du mail et donne au fichier mail son nom dfinitif */

void fusionner (char *entete, long numail, char *message)
{
    FILE *fentet, *fmess;  // descripteurs des 2 fichiers
    int  car;              // caractre du corps du mail


    // ouvrir les 2 fichiers dans le bon mode
    fentet = fopen (entete, "a");

    if (! fentet)
        // "Impossible de complter l'entte du message %s"
        errfatale ("IMPOS_MAJ_ENTETE", entete);

    fmess = fopen (message, "r");

    if (! fmess)
        // "Impossible de relire le corps du message %s"
        errfatale ("IMPOS_LIRE_MSG", message);

    // saut de ligne obligatoire entre l'entte et le message
    putc ('\n', fentet);

    // copie des donnes (on rajoute le message  l'entte)
    car = getc (fmess);

    while (car != EOF)
    {
        putc (car, fentet);
        car = getc (fmess);
    }

    // fermer les fichiers
    fclose (fmess);
    fclose (fentet);

    // on peut dtruire le fichier temporaire qui contenait le message
    unlink (message);

    // donner au fichier mail son nom dfinitif
    sprintf (message, "%s/e%07ld", direnv, numail);
    rename (entete, message);
}



/* ajoute la signature ventuelle  un mail */

void ajout_sign (char *message)
{
    char fic_sign [szchemin]; // fichier contenant la signature de l'expditeur
    char *car_fic_sign;       // pointeur sur un caractre de ce fichier
    FILE *fmess, *fsign;      // descripteur des fichiers message et signature
    int  car, carsuiv;        // caractres de la signature
    int  suiv_lu = 0;         // indique si on a lu le caractre suivant


    // fabriquer le nom du fichier signature
    strcpy (fic_sign, direnv);
    car_fic_sign = fic_sign + strlen (direnv) - strlen (ficdir ("DIR_SORTIE"));
    strcpy (car_fic_sign, ficdir ("FIC_SIGNATURE"));

    // essayer d'ouvrir ce fichier en lecture
    fsign = fopen (fic_sign, "r");

    // si le fichier signature a pu tre ouvert
    if (fsign)
    {
        // ouvrir le fichier contenant le mail en criture fin de fichier
        fmess = fopen (message, "a");

        // termin si problme  l'ouverture
        if (!fmess)
        {
            // "Fichier mail protg en criture"
            affiche_err ("ACCES_FICHMAIL_ECR");
            montecurs ();
            attendre (2);
            return;
        }

        // rajout d'une sparation avant la signature
        fputs ("--------------------------------\n", fmess);

        // copie de la signature
        car = getc (fsign);

        // si on travaille avec le jeu de caractres UTF-8
        if (util_utf8 ())
        {
            // tant que non fin de fichier
            while (car != EOF)
            {
                // si le caractre lu est un caractre ASCII
                if ((car & 0x80) == 0)
                    // recopier ce caractre
                    putc (car, fmess);
                // sinon si ce caractre est dans l'intervalle 0x80 - 0xBF
                else if ((car & 0x40) == 0)
                {
                    // gnrer la squence UTF-8 correspondante
                    putc (0xC2, fmess);
                    putc (car, fmess);
                }
                // sinon a peut tre le dbut d'une squence UTF-8
                else
                {
                    // lire le caractre suivant
                    carsuiv = getc (fsign);

                    // si c'est le 2me caractre d'une squence UTF-8
                    if ((carsuiv & 0xC0) == 0x80)
                    {
                        // recopier la squence UTF-8
                        putc (car, fmess);
                        putc (carsuiv, fmess);

                        // si elle n'est pas finie
                        if ((car & 0xE0) == 0xE0)
                        {
                            // lire et recopier la fin de la squence
                            // sans la tester
                            putc (getc (fsign), fmess);

                            if ((car & 0xF0) == 0xF0)
                                putc (getc (fsign), fmess);
                        }
                    }
                    // sinon
                    else
                    {
                        // convertir le premier caractre en UTF-8
                        putc (0xC3, fmess);
                        putc (car - 0x40, fmess);

                        // le caractre suivant est dj lu
                        car = carsuiv;
                        suiv_lu = 1;
                    }
                }

                // lire le caractre suivant si ncessaire
                if (suiv_lu)
                    suiv_lu = 0;
                else
                    car = getc (fsign);
            }
        }
        // sinon on travaille avec un jeu de caractres ISO-8859
        else
        {
            // tant que non fin de fichier
            while (car != EOF)
            {
                // si le caractre lu n'est pas le dbut d'une
                // squence UTF-8 qu'on peut convertir en ISO-8859-1
                if (car != 0xC2 && car != 0xC3)
                    // recopier le caractre
                    putc (car, fmess);
                // sinon
                else
                {
                    // lire le caractre suivant
                    carsuiv = getc (fsign);

                    // si c'est le 2me caractre d'une squence UTF-8
                    if ((carsuiv & 0xC0) == 0x80)
                    {
                        // afficher le caractre ISO-8859-1 correspondant
                        if (car == 0xC2)
                            putc (carsuiv, fmess);
                        else
                            putc (carsuiv | 0x40, fmess);
                    }
                    // sinon
                    else
                    {
                        // recopier le premier caractre
                        putc (car, fmess);

                        // le suivant est dj lu
                        car = carsuiv;
                        suiv_lu = 1;
                    }
                }

                // lire le caractre suivant si ncessaire
                if (suiv_lu)
                    suiv_lu = 0;
                else
                    car = getc (fsign);
            }
        }

        // fermer les fichiers
        fclose (fmess);
        fclose (fsign);
    }

    // sinon si le fichier signature existait quand mme
    else if (access (fic_sign, 0) == 0)
    {
        // message d'avertissement
        // "Fichier signature %s protg en lecture"
        aff_err_arg ("FICH_SIGN_PROTLECT", fic_sign);
        montecurs ();
        attendre (2);
    }
}



/* charge en mmoire le contenu du carnet d'adresse */

void charge_carnet_adr ()
{
    char ficadr [szchemin];  // fichier contenant le carnet d'adresse
    FILE *fadr;              // descripteur du fichier carnet d'adresse
    char buf_adr [szbufadr]; // buffer pour lecture d'une adresse du carnet
    char *ligne_adr;         // ligne du carnet d'adresse  mmoriser
    int  i, taille;          // compteur + longueur ligne  mmoriser


    // initialisation : aucune adresse encore mmorise
    nb_lignes = 0;

    // trouver le nom du fichier carnet d'adresse et l'ouvrir en lecture
    strcpy (ficadr, direnv);
    strcpy (ficadr + strlen (direnv) - 6, ficdir ("FIC_CARNET-ADR"));
    fadr = fopen (ficadr, "r");

    // si ce carnet d'adresse existe
    if (fadr)
    {
        // compter le nombre de lignes qu'il contient
        while (fgets (buf_adr, szbufadr, fadr))
            nb_lignes ++;

        // allouer la zone mmoire pour mmoriser ces lignes
        // la taille de la liste alloue est optimale (sauf si lignes
        // sans adresse Email valide) et non limite  la compilation
        adresse = malloc (nb_lignes * sizeof (char *));

        // et la liste des lignes slectionnes par l'utilisateur
        selection = malloc (nb_lignes);

        if ((! adresse) || (! selection))
            // "Manque de place mmoire, l'application ne peut fonctionner"
            errfatale ("MANQUE_MEMOIRE", NULL);

        // on revient au dbut du fichier
        rewind (fadr);

        // rinitialisation du compteur de lignes
        // on ne mmorisera que les lignes avec un @ (adresse Email valide)
        nb_lignes = 0;

        // rcuprer et mmoriser les diffrentes lignes du carnet d'adresses
        while (fgets (buf_adr, szbufadr, fadr))
        {
            taille = strlen (buf_adr) - 1;

            // supprimer les blancs ou parasites en fin de ligne
            while (taille > 0 && buf_adr [taille] <= ' ')
                taille --;

            // tester la validit de l'adresse
            for (i = 0; i < taille; i++)
                if (buf_adr [i] == '@')
                    break;

            // si adresse valide
            if (buf_adr [i] == '@')
            {
                // allouer un buffer pour conserver la ligne
                ligne_adr = malloc (taille + 2);

                if (ligne_adr)
                {
                    // mmoriser l'entre du carnet d'adresse
                    memcpy (ligne_adr, buf_adr, taille + 1);
                    ligne_adr [taille + 1] = '\0';
                    adresse [nb_lignes ++] = ligne_adr;
                }
                else
                // "Manque de place mmoire, l'application ne peut fonctionner"
                    errfatale ("MANQUE_MEMOIRE", NULL);
            }
        }

        // termin avec le fichier carnet d'adresse
        fclose (fadr);
    }
}



/* rcupre le nom et l'adresse de l'expditeur */

void lire_exped ()
{
    char ficadr_exped [szchemin]; // fichier contenant l'adresse de l'expditeur
    FILE *fadr;                // descripteur du fichier adresse de l'expditeur
    int  i;                    // compteur


    // dtermine le nom du fichier contenant l'adresse de l'expditeur
    strcpy (ficadr_exped, direnv);
    strcpy (ficadr_exped + strlen (direnv) - 6, ficdir ("FIC_ADR_EXPED"));

    // ouvrir ce fichier en lecture
    fadr = fopen (ficadr_exped, "r");

    // initialisation du buffer de lecture
    *buf_lect = '\0';

    // si l'ouverture s'est bien passe
    if (fadr)
    {
        // rcuprer l'information
        fgets (buf_lect, sz_buflect, fadr);
        fclose (fadr);
    }

    // si on n'a pas rcupr d'adresse
    if (! *buf_lect)
    {
        do
        {
            // demander  l'expditeur de saisir ces informations
            // "Nom, prnom et adresse Email de l'expditeur :"
            affiche_msg ("SAIS_FICADR");
            fgets (buf_lect, sz_buflect, stdin);

            // vrifier sommairement la prsence de l'adresse email
            i = 0;

            while ((buf_lect [i] != '@') && buf_lect [i])
                i++;

            // nouvelle saisie si ncessaire
            if (buf_lect [i] != '@')
                // "Adresse Email manquante"
                affiche_msg ("ADR_MANQUANTE");
        }
        while (buf_lect [i] != '@');

        // et les mmoriser dans un fichier pour les fois suivantes
        fadr = fopen (ficadr_exped, "w");

        if (fadr)
        {
            fputs (buf_lect, fadr);
            fclose (fadr);
        }
    }

    // supprimer le '\n' dans la chaine lue
    if (buf_lect [strlen (buf_lect) - 1] == '\n')
        buf_lect [strlen (buf_lect) - 1] = '\0';
}



/* slectionne des adresses de destinataires
   en tenant compte de celles dj utilises */

void choixdest (char *motcle)
{
    char buf_adr [szbufadr]; // buffer pour saisie adresse de destinataire
    int  car;                // caractre lu au clavier
    int  i, j, erreurs;      // compteurs


    // initialisation
    ajouts = 0;

    // si le carnet contient des adresses
    if (nb_lignes)
    {
        // aucun destinataire slectionn
        for (i = 0; i < nb_lignes; i++)
            selection [i] = 0;

        // initialisation affichage
        lignecour = 0;
        lignecran = 1;
        erreurs = 0;
        affpage ();

        do
        {
            // lire et traiter une touche du clavier
            car = leccar ();

            switch (car)
            {                  // dplacement dans la liste des destinataires
                case MONTE   : monte (1);
                               erreurs = 0;
                               break;

                case DESCEND : descend (1);
                               erreurs = 0;
                               break;

                case PAGEUP  : monte (lignepage + lignecran - 2);
                               erreurs = 0;
                               break;

                case PAGEDOWN: descend (2 * lignepage - lignecran - 2);
                               erreurs = 0;
                               break;

                case HOME    :
                case HOMEg   : monte (lignecour);
                               erreurs = 0;
                               break;

                case FIN     :
                case FINg    : descend (nb_lignes - lignecour - 1);
                               erreurs = 0;
                               break;

                               // slectionner une adresse
                case INSERT  : selection [lignecour] = 1;
                               affligne (lignecour);
                               break;

                               // dselectionner une adresse
                case SUPR    : selection [lignecour] = 0;
                               affligne (lignecour);
                               break;

                case ' '     : // slectionner ou dslectionner une adresse
                case '\n'    : selection [lignecour] = 1 - selection[lignecour];
                               affligne (lignecour);
                               break;

                case 0       : // aide si trop d'erreurs ou  la demande
                case F1      : effpage ();
                               // "Touches utilisables :\n"
                               // "flches Pageup, Pagedown, Home et Fin"
                               // "pour se dplacer d'une ou plusieurs lignes"
                               // "Inser pour rajouter le destinataire courant"
                               // "Suppr pour le dslectionner"
                               // "Espace ou Entre pour slect / dslect."
                               // "Control L pour rafficher la page"
                               affiche_msg ("AIDE_CHOIX_ADR-1");
                               affiche_msg ("AIDE_CHOIX_ADR-2");
                               affiche_msg ("AIDE_CHOIX_ADR-3");
                               affiche_msg ("AIDE_CHOIX_ADR-4");
                               affiche_msg ("AIDE_CHOIX_ADR-5");
                               affiche_msg ("AIDE_CHOIX_ADR-6");
                               affiche_msg ("AIDE_CHOIX_ADR-7");

                               // variante pour la touche Esc
                               if (util_systemd ())
                                   // "v ou Esc (2 fois) pour Valider la slection\n"
                                   affiche_msg ("AIDE_CHOIX_ADR-8b");
                               else
                                   // "v ou Esc pour Valider la slection\n"
                                   affiche_msg ("AIDE_CHOIX_ADR-8");

                               // "Appuyer sur une touche pour continuer"
                               affiche_msg ("ATTENTE_CLAVIER");
                               leccar ();
                               // pas de break, on continue sur affpage

                case CTRL    : affpage ();
                               erreurs = 0;

                case 'v'     : // valider la slection
                case ESC     : break;

                default      : putchar (7);  // bip

                               // si trop d'erreurs, afficher l'aide
                               if (++erreurs == 5)
                                   ungetc (0, stdin);  // on ira sur l'aide
            }
        }
        while (car != 'v' && car != ESC);

        // On va recopier les adresses slectionnes
        // et les supprimer de la liste
        j = 0;

        for (i = 0; i < nb_lignes; i++)
        {
            // si adresse slectionne
            if (selection [i])
            {
                // ajoute ce destinataire dans l'entte du fichier mail
                ajout_adr (adresse [i], motcle);

                // on peut librer la mmoire occupe par l'adresse utilise
                free (adresse [i]);
            }
            // sinon, on la conserve
            else
                adresse [j++] = adresse [i];
        }

        // mise  jour du nombre d'adresses encore utilisables
        nb_lignes = j;

        // saisie manuelle d'autres descriptions d'utilisateurs
        effpage ();

   // "Vous pouvez prciser des destinataires inconnus dans le carnet d'adresse"
        affiche_msg ("AIDE_CHOIX_ADR-9");
    }
    else
        // "Carnet d'adresse vide, vous pouvez choisir d'autres destinataires"
        affiche_msg ("AIDE_CHOIX_ADR-10");

    // "Taper juste sur entre quand il n'y en a plus"
    affiche_msg ("AIDE_CHOIX_ADR-11");

    // on repasse en mode saisie normale
    mode_normal ();

    do
    {
        // "Autre destinataire : "
        affiche_msg_nocr ("AJOUT_DEST");
        i = 0;

        do
        {
            car = getchar ();

            // on ne mmorise pas le passage  la ligne
            if (car != '\n')
                buf_adr [i++] = car;
        }
        while (car != '\n');

        // on termine la chaine lue
        buf_adr [i] = '\0';

        // et si elle n'est pas vide, on mmorise le destinataire correspondant
        if (*buf_adr)
            ajout_adr (buf_adr, motcle);
    }
    while (*buf_adr);

    // configurer la liaison clavier pour lecture directe avec timeout
    mode_raw ();

    // terminaison de la liste de destinataires de la catgorie traite
    // on ne gnre pas de passage  la ligne si aucun
    // destinataire n'a t slectionn pour cette catgorie
    if (ajouts)
        fputc ('\n', fdest);
}



/* affiche une ligne de la liste des destinataires */

void affligne (int numlig)
{
    int i;    // simple compteur

    i = 0;

    // aller en dbut de ligne
    putchar ('\r');

    // si le destinataire de la ligne  afficher a t slectionn
    if (selection [numlig])
    {
        // passer en surbrillance
        clair ();

        // mettre aussi un > en dbut de ligne (pour les terminaux braille)
        putchar ('>');
        putchar (' ');

        // afficher le destinataire
        while (adresse [numlig][i] && i < colonpage - 2)
            putchar (adresse [numlig][i++]);

        // revenir en lunminosit normale
        lumnor ();
    }
    // sinon
    else
        // afficher juste le destinataire
        while (adresse [numlig][i] && i < colonpage)
            putchar (adresse [numlig][i++]);

    // aller en dbut de ligne
    putchar ('\r');
}



/* ajoute un destinataire dans l'entte du fichier mail
   on remet en forme les informations contenues dans la chaine destin
   pour les prsenter sous la forme : "description" <adresse>
   seule l'adresse est conserve por les destinataires en copie cache */

void ajout_adr (char *destin, char *motcle)
{
    int  debnom, finom, debadr, finadr; // position infos dans la chaine destin
    char buf_adr [szbufadr];  // buffer pour mise en forme de l'adresse
    int  szbuf;        // nombre de caractres mmoriss dans buf_adr
    static int occupe; // nombre de caractres dans ligne courante du fichier
    int  i;            // simple compteur (pour copie propre des donnes)

    // on va chercher le dbut et la fin de l'adresse mail
    finadr = strlen (destin) - 1;

    // recherche de la fin de l'adresse email
    while (finadr > 0 && (destin [finadr] == ' ' || destin [finadr] == '>'
                                                 || destin [finadr] == '\t'))
        finadr --;

    // recherche du dbut de l'adresse email
    debadr = finadr - 1;

    while (debadr >= 0 && destin [debadr] != ' ' && destin [debadr] != '<'
                                                 && destin [debadr] != '\t')
        debadr --;

    debadr ++;
    szbuf = 0;

    // si l'on ne s'occupe pas des destinataires en copie cache
    if (*motcle != 'B')
    {
        // recherche de la fin du nom du destinataire
        finom = debadr - 2;

        while (finom > 0 && (destin [finom] == ' ' || destin [finom] == '"'
                                                   || destin [finom] == '\t'))
            finom --;

        // recherche du dbut du nom du destinataire
        debnom = 0;

        while (destin [debnom] == ' ' || destin [debnom] == '"'
                                      || destin [debnom] == '\t')
            debnom ++;

        // si la description du destinataire est prsente
        if (debnom < finom)
        {
            // la copier
            buf_adr [szbuf++] = '"';

            // l'utilisation de la variable i permet de ne pas
            // modifier debnom qui pourtant ne resservira plus
            for (i = debnom; i <= finom; i++)
                buf_adr [szbuf++] = destin [i];

            buf_adr [szbuf++] = '"';
            buf_adr [szbuf++] = ' ';
        }
    }

    // copie de l'adresse email
    buf_adr [szbuf++] = '<';

    // en caractres minuscules
    for (i = debadr; i <= finadr; i++)
        buf_adr [szbuf++] = tolower (destin [i]);

    buf_adr [szbuf++] = '>';
    buf_adr [szbuf++] = '\0';

    // tape 2 : on copie les informations sur le destinataire dans le fichier

    // si c'est le premier destinataire de cette catgorie
    if (! ajouts)
    {
        // on commence la liste
        fprintf (fdest, "%s:", motcle);

        // initialisations
        occupe = strlen (motcle);
        ajouts = 1;
    }
    // sinon, mettre une virgule aprs le destinataire prcdent
    else
        fputc (',', fdest);

    // passage  la ligne ?
    if (occupe + strlen (buf_adr) >= szbufadr && occupe > 4)
    {
        fprintf (fdest, "\n   ");
        occupe = 3;
    }

    // copie de la description du destinataire
    fprintf (fdest, " %s", buf_adr);

    // mise  jour du compteur d'octets dans la ligne
    occupe = occupe + strlen (buf_adr) + 2;
}
