/*
    Fichier trsfmail.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-de.txt , licence-en.txt , licence-es.txt ,
    licence-it.txt , licence-nl.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://libremail.tuxfamily.org/voir.php?page=droits


    Ce programme transfre le fichier mail pass en paramtre
    Le chemin d'accs au fichier peut tre absolu ou relatif.

    On peut choisir les destinataires  partir du carnet d'adresses
    ou en saisir de nouveaux

    Le fichier obtenu est ensuite stoqu dans le rpertoire des mails  envoyer
    Pour que cela soit possible, il faut :
    - soit que la commande trsfmail soit lance depuis la chaine vmail...
    - soit prciser le nom du rpertoire de l'arborescence des mails
*/

#define appli   // pour la dclaration de variables globales  l'application

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "messages.h"
#include "buflect.h"
#include "fmail.h"
#include "encodage.h"
#include "trtentete.h"
#include "trtligne.h"
#include "groupeligne.h"
#include "trtbordure.h"
#include "trtsection.h"
#include "numail.h"
#include "genentete.h"
#include "modepage.h"
#include "carspe.h"


/* prototypes */
void gene_entete (char *nomfic);
void ajout_entete (char *nomfic);
void copchamp (long pos_deblig);
void ajout_message ();
void copie_texte ();
void genligne (char *buffer);


/* variable globale au source
   (pour viter des passages de paramtres) */

long numail;  // numro que prendra le mail gnr



/* programme principal */

int main (int nbarg, char *varg[])
{
    char fictmp [szchemin];   // fichier de travail pour fabriquer le mail
    char ficmail [szchemin];  // nom du fichier mail dfinitif
    char *commande = ficmail; // rutilisation tableau pour appel de l'diteur
    char *editeur;            // diteur utilis pour modifier le mail


    // rcupration du nom de l'excutable
    memcom (*varg);

    // si nombre d'arguments correct
    if (--nbarg == 1 || nbarg == 2)
    {
        // ouvrir le fichier mail auquel on doit rpondre
        fmail = fopen (varg [1], "r");

        // termin si le fichier n'a pas pu tre ouvert
        if (! fmail)
        {
            if (access (varg [1], 0) == 0)
                // "Problme d'accs au fichier %s"
                errfatale ("ACCES_FICHIER", varg [1]);
            else
                // "Fichier %s inexistant"
                errfatale ("FICH_INEXISTANT", varg [1]);
        }
    }
    // sinon, rappel de la syntaxe de la commande
    else
        // "Syntaxe : %s nom_fichier_mail [rpertoire_emails]"
        psyntaxe ("SYNT_REP/TRSF-MAIL");

    // on dmarre avec la configuration standard du clavier
    mode_normal ();

    // rcuprer le nom du rpertoire des mails  envoyer
    recupdirenv (nbarg, varg [2]);

    // crer un nom de fichier temporaire pour le mail
    // il a toujours le mme nom pour viter de polluer le
    // rpertoire d'envoi en cas de fausse manoeuvre oprateur
    sprintf (fictmp, "%s/tmp-mes", direnv);

    // charger le carnet d'adresses
    charge_carnet_adr ();

    // configurer la liaison clavier pour lecture directe avec timeout
    mode_raw ();

    // fabriquer l'entte du mail de rponse
    gene_entete (fictmp);

    // retour  la configuration standard du clavier
    mode_normal ();

    // mmoriser les dates et heure d'envoi du message
    ajout_date (fictmp);

    // recopier le contenu du mail  transfrer
    ajout_entete (fictmp);

    // modification avant envoi ?
    // "Voulez vous modifier le fichier mail avant envoi ? "
    affiche_msg_nocr ("MAJ_MAIL_PRET");

    if (tolower (getchar ()) != 'n')
    {
        // si oui, choix de l'diteur de textes
        editeur = getenv ("EDITOR");

        if (! editeur)
            editeur = "vi";

        // et modification du mail
        sprintf (commande, "%s %s", editeur, fictmp);
        system (commande);
    }

    // donner au fichier mail son nom dfinitif
    sprintf (ficmail, "%s/e%07ld", direnv, numail);
    rename (fictmp, ficmail);

    // mettre  jour le fichier numail du rpertoire de sortie
    sauv_num_dermail (direnv, numail);

    // pour faire plaisir  gcc qui veut une fonction main de type int
    return (0);
}



/* gnre l'entte du mail que l'on transfre */

void gene_entete (char *nomfic)
{
    long pos_deblig;  // position dans le fichier en dbut de ligne
    // position dans le fichier mail de 2 champs de l'entte
    long posSubject, posContent;


    // fabriquer et ouvrir en criture le fichier d'entte
    fdest = fopen (nomfic, "w");

    if (! fdest)
        // "Impossible de crer un fichier d'entte %s"
        errfatale ("IMPOS_CRE_ENTETE", nomfic);

    // initialisation
    posSubject = -1;
    posContent = -1;

    // rcuprer l'adresse de l'expditeur
    lire_exped ();

    // chercher le numro du mail que l'on va crer
    numail = num_dermail (direnv) + 1;

    // gnration du champ Message-Id
    gen_mes_id (fdest, numail, buf_lect);

    // y copier l'expditeur du message
    ajouts = 0;
    ajout_adr (buf_lect, "From");
    fputc ('\n', fdest);

    // Choix des destinataires
    debpage ();
    effpage ();

    // "Destinataires principaux"
    affiche_msg_nocr ("PRINCIP_DEST");
    attendre (1);
    choixdest ("To");

    effpage ();
    // "Voulez vous mettre des destinataires en copie ? "
    affiche_msg_nocr ("ACCEPT_DEST_CC");

    if (tolower (leccar ()) != 'n')
    {
        putchar ('\n');
        choixdest ("Cc");
    }
    else
        putchar ('\n');

    effpage ();

    // "Voulez vous mettre des destinataires en copie cache ? "
    affiche_msg_nocr ("ACCEPT_DEST_BCC");

    if (tolower (leccar ()) != 'n')
    {
        putchar ('\n');
        choixdest ("Bcc");
    }

    putchar ('\n');
    effpage ();

    // Rcupration position du sujet et du type du message initial
    do
    {
        // rcuprer la position en dbut de ligne
        pos_deblig = ftell (fmail);

        // lire une ligne de l'entte du message
        lire_fmail ();

        // reprage des champs importants et mmorisation de leur position
        if (start ("Subject"))
            posSubject = pos_deblig;
        else if (start ("Content-Type"))
            posContent = pos_deblig;
    }
    while (buf_lect [0] != '\0');  // lecture entte termine si ligne vide

    // gnration du champ Subject
    if (posSubject >= 0)
    {
        // lire la ligne contenant le sujet
        fseek (fmail, posSubject, SEEK_SET);
        lire_fmail ();

        // convertir les caractres spciaux ventuels
        // c'est dans envmail qu'on rencodera ventuellement ce champ
        majlignentete ();

        // si la ligne du sujet n'est pas vide
        if (strlen (buf_lect) > 9)
        {
            // rajouter Tr: ou remplacer le Fw: dans le sujet si ncessaire
            if (buf_lect [11] == ':' && tolower (buf_lect [9]) == 't'
                                     && tolower (buf_lect [10]) == 'r')
                fputs (buf_lect, fdest);
            else if (buf_lect [11] == ':' && tolower (buf_lect [9]) == 'f'
                                          && tolower (buf_lect [10]) == 'w')
                fprintf (fdest, "Subject: Tr: %s", buf_lect + 12);
            else
                fprintf (fdest, "Subject: Tr: %s", buf_lect + 9);

            // lire la ligne suivante
            lire_fmail ();
        }
        else
        {
            // on va vrifier si le sujet est sur la ligne suivante
            lire_fmail ();

            // si c'est le cas
            if (*buf_lect == ' ' || *buf_lect == '\t')
                // gnrer le dbut du champ sujet
                fputs ("Subject: Tr: ", fdest);
            else
                // sinon, on fera comme si la mot cl Subject n'a pas t trouv
                posSubject = 0;
        }

        // tantque c'est aussi une ligne du sujet
        while (*buf_lect == ' ' || *buf_lect == '\t')
        {
            // convertir les caractres spciaux ventuels
            majlignentete ();

            // rajouter le contenu de la ligne au sujet
            fputs (buf_lect + 1, fdest);

            // et lire la ligne suivante
            lire_fmail ();
        }
    }

    // si champ sujet absent ou trouv vide lors du test prcdent
    if (posSubject == 0)
    {
        // on va en demander un  l'expditeur
        // "Sujet du mail ? "
        affiche_msg_nocr ("SUJET_MAIL");

        // retour  la configuration standard du clavier
        mode_normal ();

        // lecture clavier en tenant compte des dfauts de la commande fgets
        buf_lect [60] = '\0';
        fgets (buf_lect, 60, stdin);

        if (buf_lect [strlen (buf_lect) - 1] != '\n')
            buf_lect [strlen (buf_lect) - 1] = '\n';

        // gnrer la ligne contenant le sujet du mail (elle contient un \n)
        fprintf (fdest, "Subject: Tr: %s", buf_lect);
    }
    // sinon terminer la ligne contenant le sujet
    else
        fputc ('\n', fdest);

    // rcuprer le type principal du message et ses caractristiques
    recup_typeorig (posContent);

    // si fichier mail de type multipart/mixed
    if (ctype == MultipMixed)
    {
        // gnrer le champ Content-Type
        fprintf (fdest, "Content-Type: multipart/mixed;\n");
        fprintf (fdest, "\tboundary=\"%s\"\n", bordure [0]);
    }
    else
    {
        // sinon, le mail transfr se limitera  une zone texte
        // jeu de caractres du message
        fprintf (fdest, "Content-Type: text/plain; charset=");

        if (util_utf8 ())
            fprintf (fdest, "UTF-8;\n");
        else
            fprintf (fdest, "ISO-8859-15;\n");

        // mode d'encodage
        fprintf (fdest, "Content-Transfer-Encoding: 8bit\n");
    }

    // un copyright qui n'est pas celui d'Outlook ;-))
    // "libremail : logiciel libre multilingue"
    fprintf (fdest, "User-Agent: libremail : %s\n",
                     message ("QUALIF_LIBREMAIL"));

    // permettra de relire ce fichier
    fclose (fdest);
}



/* recopie le contenu du mail  transfrer */

void ajout_entete (char *nomfic)
{
    long pos_deblig;    // position dans le fichier en dbut de ligne
    // position dans le fichier mail des principaux champs de l'entte
    long posDate, posFrom, posXorig, posTo, posCc, posReply, posSubject;
    char *varenv_xorig; // variable d'environnement libremail_xorig
    char xorig;         // information utile de libremail_xorig


    // fabriquer et ouvrir en criture le fichier mail
    fdest = fopen (nomfic, "a");

    if (! fdest)
        // "Impossible de complter le fichier mail %s"
        errfatale ("IMPOS_MAJ_FICMAIL", nomfic);

    // initialisation
    posDate    = -1;
    posFrom    = -1;
    posXorig   = -1;
    posTo      = -1;
    posCc      = -1;
    posReply   = -1;
    posSubject = -1;

    // rcuprer la partie significative de la variable libremail_xorig
    varenv_xorig = getenv ("libremail_xorig");

    if (varenv_xorig)
        xorig = tolower (*varenv_xorig);
    else
        xorig = '\0';

    // on revient au dbut du fichier mail
    rewind (fmail);

    // on rajoute une ligne blanche pour indiquer le dbut du message
    fputc ('\n', fdest);

    // si on transfre un mail avec pices jointes
    if (ctype == MultipMixed)
    {
        // crer la premire section
        fprintf (fdest, "--%s\n", bordure [0]);

        // jeu de caractres du message
        fprintf (fdest, "Content-Type: text/plain; charset=");

        if (util_utf8 ())
            fprintf (fdest, "UTF-8;\n");
        else
            fprintf (fdest, "ISO-8859-15;\n");

        // mode d'encodage
        fprintf (fdest, "Content-Transfer-Encoding: 8bit\n\n");
    }

    // lecture de l'entte et mmorisation de la position des champs importants
    do
    {
        // rcuprer la position en dbut de ligne
        pos_deblig = ftell (fmail);

        // lire une ligne de l'entte du message
        lire_fmail ();

        // reprage des champs importants et mmorisation de leur position
        if (start ("X-Mailer") || start ("User-Agent"))
            fprintf (fdest, "%s\n", buf_lect);  // copie directe
        else if (start ("Date"))
            posDate = pos_deblig;
        else if (start ("From"))
            posFrom = pos_deblig;
        else if (start ("X-Original-From") && xorig != 'n')
            posXorig = pos_deblig;
        else if (start ("To"))
            posTo = pos_deblig;
        else if (start ("Cc"))
            posCc = pos_deblig;
        else if (start ("Reply-To"))
            posReply = pos_deblig;
        else if (start ("Subject"))
            posSubject = pos_deblig;

        // on en profite pour mmoriser le mode d'encodage
        else if (start ("Content-Transfer-Encoding"))
            mem_encodage ();
    }
    while (buf_lect [0] != '\0');  // lecture entte termine si ligne vide

    // rappel des caractristiques du message auquel on rpond
    if (posDate >= 0)
        copchamp (posDate);
    if (posFrom >= 0 && (posXorig < 0 || xorig == '2'))
        copchamp (posFrom);
    if (posXorig >= 0)
        copchamp (posXorig);
    if (posTo >= 0)
        copchamp (posTo);
    if (posCc >= 0)
        copchamp (posCc);
    if (posReply >= 0)
        copchamp (posReply);
    if (posSubject >= 0)
        copchamp (posSubject);

    // revenir au dbut du corps du message
    fseek (fmail, pos_deblig, SEEK_SET);
    lire_fmail ();

    // copie du contenu du mail
    if (ctypeorig & Multipart)
        // il faudra analyser les sections
        ajout_message ();
    else
        // une simple copie du texte suffit
        copie_texte ();

    // gnration automatique termine
    fclose (fdest);
}



/* recopie un champ de l'entte d'une ou plusieurs lignes
   repr  partir de sa position dans le fichier mail */

void copchamp (long pos_deblig)
{
    // se positionner sur la ligne contenant le champ
    fseek (fmail, pos_deblig, SEEK_SET);

    // lire cette ligne
    lire_fmail ();

    // rpter
    do
    {
        // convertir et copier la ligne
        majlignentete ();
        fprintf (fdest, "%s\n", buf_lect);

        // lire la ligne suivante
        lire_fmail ();
    }
    // jusqu' la dernire ligne du champ
    while (*buf_lect == ' ' || *buf_lect == '\t');
}



/* recopie le contenu du mail  transfrer */

void ajout_message ()
{
    // mmorisation des modes multipart trouvs
    int multipmixed = 0, multipalter = 0, multiprel = 0, multiprep = 0;


    // mmoriser tous les modes multipart imbriqus
    do
    {
        switch (ctype)
        {
            case MultipMixed : multipmixed = 1;
                               break;

            case MultipAlter : multipalter = 1;
                               break;

            case MultipRep   : multiprep   = 1;
                               break;

            case MultipRel   : multiprel   = 1;
        }

        // passer  la section suivante
        prochaine_section ();

        // rcuprer ses caractristiques
        recup_infos_section ();
    }
    while ((ctype & Multipart) && lire_fmail ());

    // indiquer si le mail peut contenir des pices jointes
    if (multipmixed)
        // "Pice(s) jointes(s) probable(s)"
        genligne (message ("PJ_PROBABLE"));

    // si une section multipart/alternative a t trouve, se positionner
    // (si l'on n'y est pas) sur la section text/plain du mail
    if (multipalter)
        posit_texte ();

    // gnrer une ligne de sparation avec l'entte
    genligne ("\n");

    // si structure du mail non conforme, message d'erreur
    if (! lire_fmail ())
        // "Pas de zone texte dans ce mail !!!"
        genligne (message ("MANQUE_ZONE_TEXTE"));

    // lecture et mmorisation du corps du message
    do
    {
        // mise en forme et mmorisation de la dernire ligne lue
        majligne ();

        if (buf_lect [strlen (buf_lect) - 1] != '\n')
            regroupeligne (buf_lect);
        else
            genligne (buf_lect);

        // et lecture de la suivante
        if (! lire_fmail ())
            return;   // on sort en fin de fichier

        // si mode multipart/report, on saute les enttes de section
        if (multiprep && nbordures && surbordure ())
        {
            do
                lire_fmail ();
            while (*buf_lect);
        }
    }
    // on s'arrte en fin de fichier, ou sur la prochaine bordure
    while (nbordures == 0 || ! surbordure ());

    // si mode multipart mixed on va recopier les pices jointes
    if (ctypeorig == MultipMixed)
    {
        // on ne conserve que la bordure de premier niveau
        // (permet de sauter le texte HTML en mode multipart/alternative)
        nbordures = 1;

        // se positionner sur la prochaine section si on n'y est pas dj
        while (! surbordure () && lire_fmail ())
            ;

        // recopier les lignes concernant les pices jointes jusqu' la fin
        do
        {
            fputs (buf_lect, fdest);
            fputc ('\n', fdest);
        }
        while (lire_fmail ());
    }
}



/* lecture et copie d'un mail de type text/plain */

void copie_texte ()
{
    // si texte encod base 64
    if (encodage_texte == Base64)
        // gnrer une ligne de sparation avec l'entte
        genligne ("\n");

    // rpter
    do
    {
        // mise en forme et mmorisation de la dernire ligne lue
        majligne ();

        if (buf_lect [strlen (buf_lect) - 1] != '\n')
            regroupeligne (buf_lect);
        else
            genligne (buf_lect);

        // et lecture de la suivante
    }
    // on s'arrte en fin de fichier
    while (lire_fmail ());
}



/* recopie une ligne du texte du mail en
   passant  la ligne si elle est trop longue */

void genligne (char *buffer)
{
    char * nouvligne;
    int  taille_ligne;
    int  i;


    // rcuprer la longueur de la ligne
    taille_ligne = strlen (buffer);

    // si ligne longue
    if (taille_ligne > 80)
    {
        // on tente de la tronquer  moins de 80 car
        i = 70;

        // chercher un blanc pour passer  la ligne
        // ou un \n rsultant d'un abus de l'encodage Mime
        while (i > 0 && buffer [i] != ' ' && buffer [i] != '\n')
            i--;

        // si pas de blanc trouv
        if (i == 0)
        {
            // on en cherche un dans l'autre sens
            i = 70;

            while (i < taille_ligne && buffer [i] != ' ' && buffer [i] != '\n')
                i++;
        }

        // mmoriser la taille de la ligne tronque
        taille_ligne = i;
    }

    // recopier les donnes
    for (i = 0; i < taille_ligne; i++)
        fputc (buffer [i], fdest);

    // si la ligne a t tronque
    if (taille_ligne < strlen (buffer))
    {
        // gnrer un passage  la ligne
        fputc ('\n', fdest);

        // et traiter la suite
        genligne (buffer + taille_ligne + 1);
    }
}
