/*
    Fichier trtsection.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-de.txt , licence-en.txt , licence-es.txt ,
    licence-it.txt , licence-nl.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://libremail.tuxfamily.org/voir.php?page=droits


    Bibliothque de fonctions permettant de parcourir les diffrentes
    sections d'un mail en mode multipart, pour accder  la zone texte
    et rcuprer la liste des fichiers joints.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "messages.h"
#include "buflect.h"
#include "fmail.h"
#include "trtentete.h"
#include "trtligne.h"
#include "encodage.h"
#include "trtsection.h"
#include "trtbordure.h"


/* prototype */
void genligne (char *buffer);



/* rcuprer la valeur du content-type, et le charset ou la bordure */

void recup_typeorig (long pos_ctype)
{
    // initialisation
    nbordures = 0;

    // si un champ Content-Type figure dans l'entte du mail
    if (pos_ctype >= 0)
    {
        // se positionner sur ce champ
        fseek (fmail, pos_ctype, SEEK_SET);
        lire_fmail ();

        // rcuprer la valeur du champ
        ctypeorig = recup_ctype ();

        // si mode multipart
        if (ctypeorig & Multipart)
        {
            // mmoriser la bordure
            while (*buf_lect && ! mem_boundary ())
                lire_fmail ();
        }
        // sinon
        else
            // rcuprer le jeu de caractres utilis
            while (*buf_lect && ! lire_charset ())
                lire_fmail ();
    }
    else
        // si pas de champ Content-Type, on affichera quand mme le mail
        ctypeorig = TextPlain;

    // pour l'instant, le type principal du mail est le type courant
    ctype = ctypeorig;
}



/* avancer jusqu' la section suivante si
   on n'est pas sur un marqueur de section */

void prochaine_section ()
{
    // tant que bordure non trouve et non fin de mail
    // remarque : il est important de tester la bordure
    //            avant de passer  la ligne
    while (! surbordure () && lire_fmail ())
        ;  // chercher la bordure sur la ligne suivante

    // si bordure trouve, lire la ligne suivante
    if (buf_lect [0] == '-')
        lire_fmail ();
}



/* mmorise le type, le jeu de caractres et le mode d'encodage de la section */

void recup_infos_section ()
{
    // initialisation : type de section non trouv
    ctype = 0;

    // tant qu'on est dans l'entte de la section, on va chercher
    // des informations dans toutes les lignes sachant qu'elles
    // peuvent apparaitre dans un ordre quelconque
    while (buf_lect [0])
    {
        // type de la section
        if (start ("Content-Type"))
            ctype = recup_ctype ();

        // mode d'encodage des caractres
        else if (start ("Content-Transfer-Encoding"))
            mem_encodage ();

        // si section multipart
        if (ctype & Multipart)
            // mmoriser la bordure quand on la trouve
            mem_boundary ();
        // sinon
        else
            // rcuprer le jeu de caractres utilis quand on le trouve
            lire_charset ();

        // passer  la ligne suivante
        lire_fmail ();
    }
}



/* se positionner sur la section texte du mail (si l'on n'y est pas) */

void posit_texte ()
{
    posit_section (TextPlain);
}



/* se positionner sur la section text/html du mail (si l'on n'y est pas) */

void posit_texthtm ()
{
    posit_section (TextHtml);
}



/* se positionner (si l'on n'y est pas) sur la section text/plain
   ou text/html du mail, en fonction du paramtre  typetexte */

void posit_section (int typetexte)
{
    // tant qu'on n'a pas trouv la section contenant
    // le texte du mail et non fin de mail
    while (ctype != typetexte && lire_fmail ())
    {
        // si une bordure a t mmorise
        if (nbordures)
            // se positionner sur la section suivante
            prochaine_section ();

        // rcuprer les informations sur la section
        recup_infos_section ();
    }

    // message d'erreur si on est arriv en fin de mail
    if (! buf_lect [0])
    {
        // Problme de structure d'un mail multi section : fin de mail atteinte
        sprintf (buf_lect, " %s\n", message ("PB_STRUCT_MAIL"));
    }
}



/* lister les pices jointes */

void liste_pj ()
{
    int  posjoint;       // position du nom du fichier joint
    char nomjoint [120]; // nom du fichier joint
    int  nbext;          // nombre d'extentions du fichier
    int  derctype;       // pour mmoriser champ content-type de la section
    int  i;              // compteur


    // on ne conserve que la bordure de premier niveau
    // (permet de sauter le texte HTML en mode multipart/alternative)
    nbordures = 1;

    // rpter
    do
    {
        // se positionner sur la prochaine section si on n'y est pas dj
        prochaine_section ();

        // rinitialisation type de section
        derctype = 0;

        // chercher et mmoriser le nom du fichier joint s'il y en a un
        do
        {
            // mmoriser les sections message/rfc822 et text/html
            if (start ("Content-Type"))
                derctype = recup_ctype ();

            // chercher si la ligne contient un nom de fichier joint
            posjoint = posnomfic ();

            // si trouv hors d'une section text/html
            if (posjoint && derctype != TextHtml)
            {
                // rcuprer son nom
                // "-> Fichier joint : "
                strcpy (nomjoint, message ("FICJOINT"));

                i = strlen (nomjoint);
                nbext = 0;

                // convertir les donnes encodes comme dans une entte
                majlignentete ();

                // recopie du nom de fichier sans dborder du tableau
                while (buf_lect [posjoint] != '"' && buf_lect [posjoint])
                {
                    // sans les (horribles) blancs qu'il pourrait contenir !!!
                    // et sans dborder du tableau
                    if (buf_lect [posjoint] != ' ' && i < sizeof (nomjoint))
                        nomjoint [i++] = buf_lect [posjoint];

                    // passer au caractre suivant
                    posjoint++;

                    // on compte les . (nombre d'extentions) du fichier
                    //  partir du 2me caractre du nom
                    if (buf_lect [posjoint] == '.')
                        nbext ++;
                }

                // terminer la chaine gnre
                if (i + 1 >= sizeof (nomjoint))
                    i = sizeof(nomjoint) - 2;

                nomjoint [i]   = '\n';
                nomjoint [i+1] = '\0';
            }
        }
        while (! posjoint && lire_fmail () && buf_lect [0] != '\0');

        // si fichier joint trouv hors d'une section text/html
        if (posjoint && derctype != TextHtml)
        {
            // actuellement tous les fichiers joints avec extention peuvent
            // contenir des virus, on accepte quand mme les .exe
            if (nbext > 1 || (nbext == 1 && strcmp (nomjoint + i - 4, ".exe\n")
                             != 0 && strcmp (nomjoint + i - 4, ".EXE\n") != 0))
            {
                // se positionner sur la premire ligne du contenu du fichier
                while (buf_lect [0] != '\0')
                    lire_fmail ();

                lire_fmail ();

                // si le fichier commence par les caractres MZ
                // (on teste l'encodage base64 correspondant)
                if (buf_lect [0] == 'T' && buf_lect [1] == 'V'
                 && buf_lect [2] >= 'o' && buf_lect [2] <= 'r')
                    // il peut s'agir d'un virus MS-DOS / Windows
                    // "     *** VIRUS ? ***\n"
                    strcpy (nomjoint + i, message ("VIRUS_POSSIBLE"));
            }

            // mmoriser le fichier joint trouv
            genligne (nomjoint);
        }

        // sinon, si section message/rfc822 trouve
        else if (derctype == Mesrfc822)
            // la rcuprer et la traiter
            recup_rfc822 ();
    }
    // jusqu'en fin de mail ou sortie de la section de premier niveau
    while (nbordures && lire_fmail ());
}



/* rcupre et traite le contenu d'une section rfc822 */

void recup_rfc822 ()
{
    char fichtmp [20];  // fichier pour mmoriser le contenu de la section
    char commande [54]; // commande  excuter pour afficher cette section
    char fichier2 [20]; // fichier gnr par cette commande
    FILE *ftmp;         // descripteur associ  un fichier


    // fabriquer un nom de fichier de travail
    sprintf (fichtmp, "/tmp/mail-1-%d", getpid ());

    // l'ouvrir en criture
    ftmp = fopen (fichtmp, "w");

    // termin si problme d'accs au fichier
    if (! ftmp)
    {
        // "Impossible de rcuprer un mail en pice jointe"
        genligne (message ("PB_RECUP_MAILJOINT"));
        return;
    }

    // afficher une ligne de sparation
    genligne ("\n");

    // lecture de la premire ligne non vide de la section
    lire_fmail ();

    // tant que non fin de section
    while (! surbordure ())
    {
        // mmoriser la ligne
        fputs (buf_lect, ftmp);
        fputc ('\n', ftmp);

        // et lire la suivante
        lire_fmail ();
    }

    // la section a t entirement rcupre
    fclose (ftmp);

    // fabriquer un nom de fichier de travail pour le mail aprs traitement
    sprintf (fichier2, "/tmp/mail-2-%d", getpid ());

    // gnrer et lancer la commande de rcupration du contenu de la section
    sprintf (commande, "voirfmail %s > %s", fichtmp, fichier2);
    system (commande);

    // on peut dtruire le premier fichier de travail
    unlink (fichtmp);

    // ouvrir le fichier obtenu en lecture
    ftmp = fopen (fichier2, "r");

    // termin si problme d'accs au fichier
    if (! ftmp)
    {
        // "Problme de droits d'accs, mail joint irrcuprable\n"
        genligne (message ("ACCES_MAILJOINT"));
        return;
    }

    // afficher les lignes prcdentes du mail
    fflush (stdout);

    // lire et traiter le fichier obtenu ligne par ligne
    while (fgets (buf_lect, sz_buflect, ftmp))
        genligne (buf_lect);

    // la section a t entirement traite
    fclose (ftmp);

    // on peut dtruire le 2me fichier de travail
    unlink (fichier2);
}
