/*
   Fichier modepage.h
   Auteur Bernard Chardonneau
*/

/*
   Logiciel de dmarrage du systme d'exploitation

   Les commandes de Libremail qui ncessitent des saisies au clavier
   durant leur excution avec utilisation des flches et autres touches
   de dplacement et la touche ESC en fin de saisie ne peuvent fonctionner
   aussi bien avec un dmarrage par systemd qu'avec un dmarrage par init.

   Avec systemd, il faudra appuyer 2 fois sur la touche ESC pour qu'elle
   soit prise en compte l o un seul appui suffit avec init.

   Par dfaut, Libremail est compil pour fonctionner que le systme
   d'exploitation ait t initialis par init ou par systemd. Une
   dtection est effectue au dmarrage des logiciels concerns pour
   programmer au mieux la rception de caractres du clavier.

   Il est possible d'viter cette dtection en compilant Libremail
   uniquement pour l'un des systmes d dmarrage, ce qui se fait en
   dcommentant l'un des deux #define qui suit.
*/

//#define systemd 0   // dcommenter pour compiler pour init uniquement
//#define systemd 1   // dcommenter pour compiler pour systemd uniquement


/* taille par defaut d'une fentre */

#define defaut_ligne   25
#define defaut_colon   80


/* macros de gestion de l'affichage */

#define putseq(s)   printf ("%c[%s", ESC,s) // affiche squence d'chappement

#define debpage()   putseq ("H")   // se positionne en haut de page
#define baspage()   putseq ("99B") // se positionne en bas de page
#define effpage()   putseq ("J")   // efface la page  partir du curseur
#define effligne()  putseq ("K")   // efface la ligne  partir du curseur
#define insligne()  putseq ("L")   // insre une ligne au niveau ligne courante
#define montecurs() putseq ("A")   // remonte le curseur d'une ligne
#define lumnor()    putseq ("0m")  // affichage en luminosit normale


/* commandes stty pour le mode de gestion du clavier */

#define cmd_mode_raw_init     "stty -icanon min 0 time 5 -echo"
#define cmd_mode_raw_systemd  "stty -icanon -echo"
#define cmd_mode_normal       "stty icanon echo"


/* prototypes des fonctions de la bibliothque modepage */

int  util_systemd ();
void mode_raw ();
void mode_normal ();
void clair ();
void sombre ();
void affpage ();
void monte (int n);
void descend (int n);
void lig_col ();
int  leccar ();
void execom (char *outil, char *fichier);
void attendre (int duree);


/* variables globales  l'application utilises par modepage
   (pour viter des tonnes de passages de paramtres) */

#ifdef modepage
     int nb_lignes;            // nombre de lignes mmorises
     int lignecour;            // numro de la ligne courante
     int lignepage, colonpage; // nombre ligne et colonnes de la page d'cran
     int lignecran;            // numro de ligne du curseur  l'cran
#else
     extern int nb_lignes;
     extern int lignecour;
     extern int lignepage, colonpage;
     extern int lignecran;
#endif
